@@ -9,6 +9,7 @@
 #pragma once
 
 #include <QObject>
+#include <QWidget>
 
 #include <Enumeration.h>
 
@@ -19,8 +20,8 @@ class TerminalBell : public QObject
     Q_OBJECT
 
 public:
-    TerminalBell(Enum::BellModeEnum bellMode);
-    void bell(const QString &message, bool terminalHasFocus);
+    explicit TerminalBell(Enum::BellModeEnum bellMode);
+    void bell(QWidget *td, const QString &message, bool terminalHasFocus);
 
     /**
      * Sets the type of effect used to alert the user when a 'bell' occurs in the
